///////////////////////////////////////////
//
// file:  MIX_device.cpp
//
// Copyright 2004 John R. Ashmun
//                jrashmun@halcyon.com
//
//    This file is part of MIX. 
//
//    MIX is free software; you can redistribute it and/or modify 
//    it under the terms of the GNU General Public License as
//    published by the Free Software Foundation; either version 2
//    of the License, or (at your option) any later version. 
//
//    MIX is distributed in the hope that it will be useful, 
//    but WITHOUT ANY WARRANTY; without even the implied warranty of 
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
//    GNU General Public License for more details. 
//
//    You should have received a copy of the GNU General Public
//    License along with MIX; if not, write to the
//    Free Software Foundation, Inc.
//    59 Temple Place, Suite 330
//    Boston MA 02111-1307  USA
//
///////////////////////////////////////////


#include <cstdio>

#include <OS.h>
#include <Path.h>

#include "MIX.h"


MIX_device::MIX_device( )
   : _interrupt( false ), _state( true ), _unit( -1 ), _blocksize( 0 )
{
}


MIX_device::~MIX_device( )
{
   if( _file )
      delete _file;
}


MIX_device::MIX_device( const MIX_device& M_d )
   : _interrupt( false ), _state( true ), _unit( -1 )
{
}


MIX_device& MIX_device::operator=( const MIX_device& M_d )
{
   _unit = -1;
   _state = true;

   return *this;
}


MIX_device::MIX_device( const int Unit )
   : _interrupt( false ), _state( true ), _unit( Unit )
{
   char unit_C_string[3];

   BPath path( "../MIX_device_files" );

   if( _unit < 16 )
   {
      _blocksize = 100;
      if( _unit < 8 )
         path.Append( "MIX_tape_units" );
      else
         path.Append( "MIX_disc_units" );

      if( _unit < 10 )
      {
         unit_C_string[0] = '0' + _unit;
         unit_C_string[1] = '\0';
      }
      else
      {
         unit_C_string[0] = '1';
         unit_C_string[1] = '0' + _unit - 10;
         unit_C_string[2] = '\0';
      }
      path.Append( unit_C_string );

      if( _unit < 8 )
         path.Append( "tape" );
      else
         path.Append( "disc_pack" );

      _file = new BFile( path.Path( ), B_READ_WRITE );
   }
   else if( _unit < 18 )
   {
      _blocksize = 16;

      if( _unit == 17 )
      {
         path.Append( "MIX_card_punch" );

         _file = new BFile( path.Path( ), B_WRITE_ONLY );
      }
   }
   else if( _unit < 19 )
      _blocksize = 24;
   else
      _blocksize = 14;

   if( _unit > 15 )
   {
      _state = false;
   }
   else if( _file->InitCheck( ) == B_OK )
      _state = false;
}


void MIX_device::unit( const int Unit )
{
   char unit_C_string[3];

   BPath path( "../MIX_device_files" );

   _interrupt = false;

   _unit = Unit;

   if( _unit < 16 )
   {
      _blocksize = 100;
      if( _unit < 8 )
         path.Append( "MIX_tape_units" );
      else
         path.Append( "MIX_disc_units" );

      if( _unit < 10 )
      {
         unit_C_string[0] = '0' + _unit;
         unit_C_string[1] = '\0';
      }
      else
      {
         unit_C_string[0] = '1';
         unit_C_string[1] = '0' + _unit - 10;
         unit_C_string[2] = '\0';
      }
      path.Append( unit_C_string );

      if( _unit < 8 )
         path.Append( "tape" );
      else
         path.Append( "disc_pack" );

      _file = new BFile( path.Path( ), B_READ_WRITE );
   }
   else if( _unit < 18 )
   {
      _blocksize = 16;

      if( _unit == 17 )
      {
         path.Append( "MIX_card_punch" );
         path.Append( "MIX_punched_cards.txt" );

         _file = new BFile( path.Path( ), B_WRITE_ONLY );
      }
   }
   else if( _unit < 19 )
      _blocksize = 24;
   else
   {
      _blocksize = 14;

      path.Append( "MIX_paper_tape_reader" );
      path.Append( "MIX_paper_tape.txt" );

      _file = new BFile( path.Path( ), B_READ_ONLY );
   }

   if( _unit > 15 )
   {
      _state = false;
   }
   else if( _file->InitCheck( ) == B_OK )
      _state = false;
}


bool MIX_device::busy( ) const
{
   return _state;
}


bool MIX_device::ready( ) const
{
   return !_state;
}


void MIX_device::read( int index )
{
   status_t    result = B_OK;

   MIX_IO_data data;

   data.device = this;
   data.control_state = control_state;
   data.interrupts_enabled = interrupts_enabled;
   data.index = index;

   thread_id read_thread = spawn_thread( _MIX_device_read,
                                         "MIX_read",
                                         0,
                                         0 );

   result = send_data( read_thread, 0, (void*)&data, sizeof( data ) );

   for( ; ; )
   {
      if( !_state )
         break;

      snooze( 100 );

      u_counter += 10;
   }

   _state = true;

   resume_thread( read_thread );
}


void MIX_device::write( int index )
{
   status_t    result = B_OK;

   MIX_IO_data data;

   data.device = this;
   data.control_state = control_state;
   data.interrupts_enabled = interrupts_enabled;
   data.index = index;

   thread_id write_thread = spawn_thread( _MIX_device_write,
                                          "MIX_write",
                                          0,
                                          0 );

   result = send_data( write_thread, 0, (void*)&data, sizeof( data ) );

   for( ; ; )
   {
      if( !_state )
         break;

      snooze( 100 );

      u_counter += 10;
   }

   _state = true;

   resume_thread( write_thread );
}


void MIX_device::control( int index )
{
   status_t    result = B_OK;

   MIX_IO_data data;

   data.device = this;
   data.control_state = control_state;
   data.interrupts_enabled = interrupts_enabled;
   data.index = INST.M( );

   thread_id control_thread = spawn_thread( _MIX_device_control,
                                            "MIX_control",
                                            0,
                                            0 );

   result = send_data( control_thread, 0, (void*)&data, sizeof( data ) );

   for( ; ; )
   {
      if( !_state )
         break;

      snooze( 100 );

      u_counter += 10;
   }

   _state = true;

   resume_thread( control_thread );
}


bool MIX_device::interrupt( ) const
{
   return _interrupt;
}


void MIX_device::clear_interrupt( )
{
   _interrupt = false;
}


int32 _MIX_device_read( void* )
{
   bool        interrupts_were_enabled = false;
   bool        was_in_control_state = false;

   char        buffer[72];

   int         code = 0;
   int         i = 0;
   int         index = 0;
   int         num_bytes = 0;
   int         values[100];

   off_t       block_offset = 0;

   thread_id   sender = 0;

   MIX_device* device = 0;

   MIX_IO_data data;

   code = receive_data( &sender, &data, sizeof( data ) );

   device = data.device;
   was_in_control_state = data.control_state;
   interrupts_were_enabled = data.interrupts_enabled;
   index = data.index;

   num_bytes = device->_blocksize * 5;

   if( device->_unit == 16 )
   {
      // Card reader reads the MIX character codes.
      if( (index >= 0) && ((index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            MEMORY[index].Byte_1( ascii_to_MIX( getchar( ) ) );
            i++;

            MEMORY[index].Byte_2( ascii_to_MIX( getchar( ) ) );
            i++;   

            MEMORY[index].Byte_3( ascii_to_MIX( getchar( ) ) );
            i++;   

            MEMORY[index].Byte_4( ascii_to_MIX( getchar( ) ) );
            i++;   

            MEMORY[index++].Byte_5( ascii_to_MIX( getchar( ) ) );
         }

         // Consume the | and \n characters that show the end of the
         // block in a text editor.
         getchar( );
         getchar( );

         device->_state = false;
      }
      else if( was_in_control_state && ((-index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            cs_MEMORY[-index].Byte_1( ascii_to_MIX( getchar( ) ) );
            i++;

            cs_MEMORY[-index].Byte_2( ascii_to_MIX( getchar( ) ) );
            i++;

            cs_MEMORY[-index].Byte_3( ascii_to_MIX( getchar( ) ) );
            i++;

            cs_MEMORY[-index].Byte_4( ascii_to_MIX( getchar( ) ) );
            i++;

            cs_MEMORY[-index--].Byte_5( ascii_to_MIX( getchar( ) ) );
         }

         // Consume the | and \n characters that show the end of the
         // block in a text editor.
         getchar( );
         getchar( );

         device->_state = false;
      }
   }
   else if( device->_unit == 19 )
   {
      // Paper tape reader reads the MIX character codes.
      if( (index >= 0) && ((index + num_bytes) < 4000) )
      {
         device->_file->Read( buffer, sizeof( buffer ) );

         for( i = 0; i < num_bytes; i++ )
         {
            MEMORY[index].Byte_1( ascii_to_MIX( buffer[i] ) );
            i++;

            MEMORY[index].Byte_2( ascii_to_MIX( buffer[i] ) );
            i++;   

            MEMORY[index].Byte_3( ascii_to_MIX( buffer[i] ) );
            i++;   

            MEMORY[index].Byte_4( ascii_to_MIX( buffer[i] ) );
            i++;   

            MEMORY[index++].Byte_5( ascii_to_MIX( buffer[i] ) );
         }

         device->_state = false;
      }
      else if( was_in_control_state && ((-index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            cs_MEMORY[-index].Byte_1( ascii_to_MIX( buffer[i] ) );
            i++;

            cs_MEMORY[-index].Byte_2( ascii_to_MIX( buffer[i] ) );
            i++;

            cs_MEMORY[-index].Byte_3( ascii_to_MIX( buffer[i] ) );
            i++;

            cs_MEMORY[-index].Byte_4( ascii_to_MIX( buffer[i] ) );
            i++;

            cs_MEMORY[-index--].Byte_5( ascii_to_MIX( buffer[i] ) );
         }

         device->_state = false;
      }
   }
   else if( device->_unit > 7 )
   {
      // MIX discs are positioned by the contents
      // of bytes 4 & 5 of the X register and do
      // unformatted I/O.
      if( (index >= 0) && ((index + device->_blocksize) < 4000) )
      {
         block_offset = (XREG.Byte_4( ) << 6) | XREG.Byte_5( );
         if( block_offset != device->_file->Position( ) )
            device->_file->Seek( block_offset, SEEK_SET );

         device->_file->Read( values, sizeof( values ) );

            for( i = 0; i < device->_blocksize; i++ )
               MEMORY[index++].int_value( values[i] );

         device->_state = false;
      }
      else if( was_in_control_state 
               && (index < 0)
               && ((-index + device->_blocksize) < 4000) )
      {
         block_offset = (XREG.Byte_4( ) << 6) | XREG.Byte_5( );
         if( block_offset != device->_file->Position( ) )
            device->_file->Seek( block_offset, SEEK_SET );

         device->_file->Read( values, sizeof( values ) );

         for( i = 0; i < device->_blocksize; i++ )
            cs_MEMORY[-index--].int_value( values[i] );

         device->_state = false;
      }
   }
   else
   {
      // MIX tapes do unformatted I/O.
      if( (index >= 0) && ((index + device->_blocksize) < 4000) )
      {
         device->_file->Read( values, sizeof( values ) );

            for( i = 0; i < device->_blocksize; i++ )
               MEMORY[index++].int_value( values[i] );

         device->_state = false;
      }
      else if( was_in_control_state
               && (index < 0)
               && ((-index + device->_blocksize) < 4000) )
      {
         device->_file->Read( values, sizeof( values ) );

            for( i = 0; i < device->_blocksize; i++ )
               cs_MEMORY[-index--].int_value( values[i] );

         device->_state = false;
      }
   }

   if( !device->_state && was_in_control_state && interrupts_were_enabled )
      device->_interrupt = true;

   return 0;
}


int32 _MIX_device_write( void* )
{
   bool        interrupts_were_enabled = false;
   bool        was_in_control_state = false;

   char        buffer[82];

   int         code = 0;
   int         i = 0;
   int         index = 0;
   int         num_bytes = 0;
   int         values[100];

   off_t       block_offset = 0;

   thread_id   sender = 0;

   MIX_device* device = 0;

   MIX_IO_data data;

   code = receive_data( &sender, &data, sizeof( data ) );

   device = data.device;
   was_in_control_state = data.control_state;
   interrupts_were_enabled = data.interrupts_enabled;
   index = data.index;

   num_bytes = device->_blocksize * 5;

   if( device->_unit == 17 )
   {
      buffer[80] = '|';
      buffer[81] = '\n';

      if( (index >= 0) && ((index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            buffer[i] = MIX_to_ascii( MEMORY[index].Byte_1( ) );
            i++;

            buffer[i] = MIX_to_ascii( MEMORY[index].Byte_2( ) );
            i++;

            buffer[i] = MIX_to_ascii( MEMORY[index].Byte_3( ) );
            i++;

            buffer[i] = MIX_to_ascii( MEMORY[index].Byte_4( ) );
            i++;

            buffer[i] = MIX_to_ascii( MEMORY[index++].Byte_5( ) );
         }

         device->_file->Write( buffer, sizeof( buffer ) );

         device->_state = false;
      }
      else if( was_in_control_state && (index < 0) && ((-index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            buffer[i] = MIX_to_ascii( cs_MEMORY[-index].Byte_1( ) );
            i++;

            buffer[i] = MIX_to_ascii( cs_MEMORY[-index].Byte_2( ) );
            i++;

            buffer[i] = MIX_to_ascii( cs_MEMORY[-index].Byte_3( ) );
            i++;

            buffer[i] = MIX_to_ascii( cs_MEMORY[-index].Byte_4( ) );
            i++;

            buffer[i] = MIX_to_ascii( cs_MEMORY[-index--].Byte_5( ) );
         }

         device->_file->Write( buffer, sizeof( buffer ) );

         device->_state = false;
      }
   }
   else if( device->_unit == 18 )
   {
      // Printer writes a line of
      // 120 MIX character codes.
      if( (index >= 0) && ((index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            putchar( MIX_to_ascii( MEMORY[index].Byte_1( ) ) );
            i++;

            putchar( MIX_to_ascii( MEMORY[index].Byte_2( ) ) );
            i++;   

            putchar( MIX_to_ascii( MEMORY[index].Byte_3( ) ) );
            i++;   

            putchar( MIX_to_ascii( MEMORY[index].Byte_4( ) ) );
            i++;   

            putchar( MIX_to_ascii( MEMORY[index++].Byte_5( ) ) );
         }

         putchar( '\n' );

         device->_state = false;
      }
      else if( was_in_control_state && (index < 0) && ((-index + num_bytes) < 4000) )
      {
         for( i = 0; i < num_bytes; i++ )
         {
            putchar( MIX_to_ascii( cs_MEMORY[-index].Byte_1( ) ) );
            i++;

            putchar( MIX_to_ascii( cs_MEMORY[-index].Byte_2( ) ) );
            i++;

            putchar( MIX_to_ascii( cs_MEMORY[-index].Byte_3( ) ) );
            i++;

            putchar( MIX_to_ascii( cs_MEMORY[-index].Byte_4( ) ) );
            i++;

            putchar( MIX_to_ascii( cs_MEMORY[-index--].Byte_5( ) ) );
         }

         putchar( '\n' );

         device->_state = false;
      }
   }
   else if( device->_unit > 7 )
   {
      // MIX discs are positioned by the contents
      // of bytes 4 & 5 of the X register and do
      // unformatted I/O.
      if( (index >= 0) && ((index + device->_blocksize) < 4000) )
      {
         block_offset = (XREG.Byte_4( ) << 6) | XREG.Byte_5( );
         if( block_offset != device->_file->Position( ) )
            device->_file->Seek( block_offset, SEEK_SET );

         for( i = 0; i < device->_blocksize; i++ )
            values[i] = MEMORY[index++].int_value( );

         device->_file->Write( values, sizeof( values ) );

         device->_state = false;
      }
      else if( was_in_control_state
               && (index < 0)
               && ((-index + device->_blocksize) < 4000) )
      {
         block_offset = (XREG.Byte_4( ) << 6) | XREG.Byte_5( );
         if( block_offset != device->_file->Position( ) )
            device->_file->Seek( block_offset, SEEK_SET );

         for( i = 0; i < device->_blocksize; i++ )
            values[i] = cs_MEMORY[-index--].int_value( );

         device->_file->Write( values, sizeof( values ) );

         device->_state = false;
      }
   }
   else
   {
      // MIX tapes do unformatted I/O.
      if( (index >= 0) && ((index + device->_blocksize) < 4000) )
      {
         for( i = 0; i < device->_blocksize; i++ )
            values[i] = MEMORY[index++].int_value( );

         device->_file->Write( values, sizeof( values ) );

         device->_state = false;
      }
      else if( was_in_control_state
               && (index < 0)
               && ((-index + device->_blocksize) < 4000) )
      {
         for( i = 0; i < device->_blocksize; i++ )
            values[i] = cs_MEMORY[-index--].int_value( );

         device->_file->Write( values, sizeof( values ) );

         device->_state = false;
      }
   }

   if( !device->_state && was_in_control_state && interrupts_were_enabled )
      device->_interrupt = true;

   return 0;
}


int32 _MIX_device_control( void* )
{
   bool        interrupts_were_enabled = false;
   bool        was_in_control_state = false;

   int         M = 0;
   int         code = 0;

   thread_id   sender = 0;

   off_t       block_offset = 0;
   off_t       current_offset = 0;
   off_t       size = 0;

   MIX_device* device = 0;

   MIX_IO_data data;

   code = receive_data( &sender, &data, sizeof( data ) );

   device = data.device;
   was_in_control_state = data.control_state;
   interrupts_were_enabled = data.interrupts_enabled;

   if( device->_unit < 8 )
   {
      // Position a tape.
      M = data.index;

      current_offset = device->_file->Position( );

      if( M == 0 )
      {
         // Rewind the tape.

         block_offset = 0;
         device->_file->Seek( block_offset, SEEK_SET );

         device->_state = false;
      }
      else if( M < 0 )
      {
         // Back up one block.

         if( current_offset >= 400 )
         {
            block_offset = -400;
            device->_file->Seek( block_offset, SEEK_CUR );
         }

         device->_state = false;
      }
      else
      {
         // Go forward one block.

         device->_file->GetSize( &size );
         if( (size - current_offset) >= 400 )
         {
            block_offset = 400;
            device->_file->Seek( block_offset, SEEK_CUR );

            device->_state = false;
         }

         // Note:  Attempting to position forward beyond
         //        the last record that was written is
         //        not permitted, and leaves the tape
         //        unit busy forever, or at least until
         //        MIX is restarted.
      }
   }
   else if( device->_unit < 16 )
   {
      // Disc drive seek.
      block_offset = (XREG.Byte_4( ) << 6) | XREG.Byte_5( );
      if( block_offset != device->_file->Position( ) )
         device->_file->Seek( block_offset, SEEK_SET );

      device->_state = false;
   }
   else
   {
      // Formatted devices:  just set their state to ready.

      device->_state = false;
   }

   if( !device->_state && was_in_control_state && interrupts_were_enabled )
      device->_interrupt = true;

   return 0;
}
