///////////////////////////////////////////
//
// file:  SPEC.cpp
//
// Copyright 2004 John R. Ashmun
//                jrashmun@halcyon.com
//
//    This file is part of MIX. 
//
//    MIX is free software; you can redistribute it and/or modify 
//    it under the terms of the GNU General Public License as
//    published by the Free Software Foundation; either version 2
//    of the License, or (at your option) any later version. 
//
//    MIX is distributed in the hope that it will be useful, 
//    but WITHOUT ANY WARRANTY; without even the implied warranty of 
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
//    GNU General Public License for more details. 
//
//    You should have received a copy of the GNU General Public
//    License along with MIX; if not, write to the
//    Free Software Foundation, Inc.
//    59 Temple Place, Suite 330
//    Boston MA 02111-1307  USA
//
///////////////////////////////////////////


#include <Application.h>

#include "MIX.h"
#include "Alerts.h"


int MIX_to_digit( int );
int digit_to_MIX( int );


void SPEC( )
{
   char*    text = "MIX has executed a HLT instruction.";

   int      areg_magnitude = 0;
   int      f = INST.F( );
   int      num = 0;

   switch( f )
   {
      case 0:
         num = MIX_to_digit( XREG.Byte_5( ) );
         num += 10 * MIX_to_digit( XREG.Byte_4( ) );
         num += 100 * MIX_to_digit( XREG.Byte_3( ) );
         num += 1000 * MIX_to_digit( XREG.Byte_2( ) );
         num += 10000 * MIX_to_digit( XREG.Byte_1( ) );
         num += 100000 * MIX_to_digit( AREG.Byte_5( ) );
         num += 1000000 * MIX_to_digit( AREG.Byte_4( ) );
         num += 10000000 * MIX_to_digit( AREG.Byte_3( ) );
         num += 100000000 * MIX_to_digit( AREG.Byte_2( ) );
         num += 1000000000 * MIX_to_digit( AREG.Byte_1( ) );

         AREG.Byte_1( (num >> 24) & 0x3F );
         AREG.Byte_2( (num >> 18) & 0x3F );
         AREG.Byte_3( (num >> 12) & 0x3F );
         AREG.Byte_4( (num >> 6) & 0x3F );
         AREG.Byte_5( num & 0x3F );
         break;

      case 1:
         areg_magnitude = (AREG.Byte_1( ) << 24)
                          + (AREG.Byte_2( ) << 18)
                          + (AREG.Byte_3( ) << 12)
                          + (AREG.Byte_4( ) << 6)
                          + AREG.Byte_5( );

         AREG.Byte_1( digit_to_MIX( areg_magnitude / 1000000000 ) );
         areg_magnitude -= 1000000000 * (areg_magnitude / 1000000000);
         AREG.Byte_2( digit_to_MIX( areg_magnitude / 100000000 ) );
         areg_magnitude -= 100000000 * (areg_magnitude / 100000000);
         AREG.Byte_3( digit_to_MIX( areg_magnitude / 10000000 ) );
         areg_magnitude -= 10000000 * (areg_magnitude / 10000000);
         AREG.Byte_4( digit_to_MIX( areg_magnitude / 1000000 ) );
         areg_magnitude -= 1000000 * (areg_magnitude / 1000000);
         AREG.Byte_5( digit_to_MIX( areg_magnitude / 100000 ) );
         areg_magnitude -= 100000 * (areg_magnitude / 100000);
         XREG.Byte_1( digit_to_MIX( areg_magnitude / 10000 ) );
         areg_magnitude -= 10000 * (areg_magnitude / 10000);
         XREG.Byte_2( digit_to_MIX( areg_magnitude / 1000 ) );
         areg_magnitude -= 1000 * (areg_magnitude / 1000);
         XREG.Byte_3( digit_to_MIX( areg_magnitude / 100 ) );
         areg_magnitude -= 100 * (areg_magnitude / 100);
         XREG.Byte_4( digit_to_MIX( areg_magnitude / 10 ) );
         areg_magnitude -= 10 * (areg_magnitude / 10);
         XREG.Byte_5( digit_to_MIX( areg_magnitude ) );
         break;

      case 2:
         DisplayInfo( text );

         cycle = false;

         break;

      case 3:
         interrupts_enabled = true;
         break;

      case 4:
         interrupts_enabled = false;
         break;

      case 7:
         if( !control_state )
         {
            enter_control_state( );
            program_counter = -12;
         }
         else
            enter_normal_state( );

         u_counter++;

         break;

      default:
         break;
   }

   u_counter++;
}
